from rocolib.utils.dimensions import isDim
from rocolib.utils.numsym import Dummy


PARAM_TYPES = {
    "length": {
        "valueType": "(float, int)",
        "minValue": 0,
        "units": "mm",
    },
    "angle": {
        "valueType": "(float, int)",
        "minValue": 0,
        "maxValue": 360,
        "units": "degrees",
    },
    "ratio": {
        "valueType": "(float, int)",
        "minValue": 0,
        "maxValue": 1,
    },
    "count": {
        "valueType": "int",
        "minValue": 0,
    },
    "dimension": {
        "valueType": "str",
        #"isValid": isDim,
    },
}

class Parameter:
    def __init__(self, name, defaultValue=None, paramType=None, **kwargs):
        ### XXX "." is used to separate subcomponent parameters when inherited
        ###     but we can't tell the difference here, so it's not invalid I guess

        #if "." in name:
            #raise ValueError("Invalid character '.' in parameter name " + name)
        self.name = name

        if defaultValue is None and not kwargs.get("optional", False):
            raise ValueError(f"Must specify either defaultValue or optional=True for parameter {name}")

        self.defaultValue = defaultValue
        self.spec = {}

        if paramType in PARAM_TYPES:
            for k, v in PARAM_TYPES[paramType].items():
                self.spec[k] = v
        elif paramType:
            raise ValueError(f"Unknown paramType: {paramType}")
        for k, v in kwargs.items():
            self.spec[k] = v
        self.assertValid(defaultValue)

        vt = self.spec.get("valueType", "")
        no = not(self.spec.get("optional", False))

        if no and ("int" in vt or "float" in vt):
            integer=None
            positive=None

            if "float" not in vt:
                integer=True

            mv = self.spec.get("minValue", -1)
            if mv is not None and mv >= 0:
                positive=True

            self.symbol = Dummy(name, real=True, positive=positive, integer=integer)
        else:
            self.symbol = None
        self.value = None

    def setValue(self, value):
        self.assertValid(value)
        self.value = value

    def setDefault(self, force=False):
        if force or self.value is None:
            self.setValue(self.defaultValue)

    def assertValid(self, value):
        if value is None:
            if self.spec.get("optional", False):
                self.value = value
                return
            else:
                raise ValueError(f"Parameter {self.name} is not optional and cannot be set to None")

        def check(spec, test, error):
            if (self.spec.get(spec, None) is not None and not test(self.spec[spec])):
                raise ValueError(f"When setting parameter {self.name}: {value} {error} {self.spec[spec]}")

        check("valueType", lambda x : isinstance(value, eval(x)), "is not of type")
        check("minValue", lambda x : value >= x, "is less than")
        check("maxValue", lambda x : value <= x, "is greater than")
        check("isValid", lambda x : x(value), "is invalid")
        return True

    def getValue(self):
        if self.value is not None:
            return self.value
        else:
            return self.symbol

    def getSpec(self):
        return self.spec

class Parameterized(object):
    """
    Like a dictionary k/v store, but we require special syntax constructs
    to set/update keys

    XXX FIX: Name is duplicated both in the Parameter object and the dict key
    """
    def __init__(self):
        self.parameters = {}


    def addParameter(self, name, defaultValue=None, **kwargs):
        """
        Adds a k/v pair to the internal store if the key has not been added before
        Raises KeyError if the key has been added before
        """
        if name in self.parameters:
            raise KeyError("Parameter %s already exists on object %s" % (name, str(self)))
        p = Parameter(name, defaultValue, **kwargs)
        self.parameters.setdefault(name, p)
        return p


    def useDefaultParameters(self, force=False):
        """
        Assign the default value to all parameters:
        - Only if otherwise unset if force=False
        - Overwrite any previously set values if force=True
        """
        for n, p in self.parameters.items():
            p.setDefault(force)


    def setParameter(self, n, v):
        """
        Sets a k/v pair to the internal store if the key has been added previously
        Raises KeyError if the key has not been added before
        Passes along any ValueErrors from Parameter object
        """
        if n in self.parameters:
            self.parameters[n].setValue(v)
        else:
            raise KeyError("Parameter %s not initialized on object %s" % (n, str(self)))


    def getParameter(self, name):
        """
        Retrieves the parameter value with the given name
        Raises KeyError if the key is not been set
        """
        return self.parameters[name].getValue()


    def getParameterInfo(self):
        """
        Retrieves the parameter metadata info
        """
        return {k: v.__dict__ for k, v in self.parameters.items()}


    def hasParameter(self, name):
        return name in self.parameters


    def delParameter(self, name):
        self.parameters.pop(name)

